# ======================================================
# ABOUT
# ===========================
'''
manilib - Alpha 0.1 - by PitBull0993

The only thing I did test was the clients.txt-parsers as well as the addon management.
There is a chance of about 0.1% that you do not encounter a bug.
If you catch a bug, report it on the Mattie-Forums or by email (pitkley@gmail.com).

The manilib was written by me (PitBull0993) in about 5 months. But there were a lot of
breaks, so the actual time would be about 1 to 2 months. When I wrote the clients.txt-parser,
I only had a Python Shell and Idle (I were on vacation). When I came back, I stumbled a
few days later upon the keyvalues-lib. With this library, you could code a fully
working parser in about 6 to 20 lines I guess. Well, I did it in 146 lines =D. In my tests,
the parser worked without any problems. But if you encounter any errors, I will either fix it,
or completely recode the parser with keyvalues. For now I will let it like it is.
The idea as well as some code parts for the addon management, I got by the gungamelib. The
GG5-developers did a really great job, and I like how they handle the custom addons.
The addon management of manilib is not as extensive as the one of GunGame, but in my eyes
it's exactly what is needed by custom addons.
'''
# ======================================================


# ======================================================


# ======================================================
# IMPORTS
# ===========================
import es
import os
import cPickle
# ======================================================


# ======================================================
# VARS
# ===========================
dict_mani = {}
dict_caddons = {}

gamePath = str(es.ServerVar('eventscripts_gamedir'))
gamePath = gamePath.replace('\\', '/')
clientsPath = "%s/cfg/mani_admin_plugin/clients.txt" % gamePath
cflagsPath = "%s/addons/eventscripts/mani/data/cflags.db" % gamePath

aflags = {'a':'Private Admin Chat', 'A':'Accept Vote', 'admin':'Basic Admin', 'b':'non-permanent ban',
          'B':'RCON voting menu', 'c':'Map Change', 'C':'Cancel Vote', 'client':'Create Clients and Set Powers',
          'd':'Drug', 'D':'Menu Question Vote', 'e':'Explode', 'E':'View client rates', 'f':'Freeze', 'F':'Burn',
          'g':'Swap Player', 'G':'No Clip Mode', 'grav':'Adjust Player Gravity', 'H':'War Mode', 'i':'Blind', 'I':'Mute',
          'j':'Gimp', 'J':'Reset Stats', 'k':'Kick', 'K':'Cash', 'l':'Slap', 'L':'RCON say', 'm':'Slay', 'M':'Admin Skins',
          'N':'Set skins', 'o':'Admin Chat', 'O':'Drop C4', 'p':'Sound', 'P':'Set Client Flags', 'pban':'Permanent Ban',
          'q':'RCON menu level 1', 'Q':'Console Question Vote', 'q2':'RCON menu level 2', 'q3':'RCON menu level 3', 'r':'RCON',
		  'R':'RCON voting', 'reserved':'Reserved Slot', 's':'Admin Say', 'S':'Set Skin color', 'spray':'Spray tag tracking', 't':'Teleport',
          'T':'Time Bomb', 'U':'Fire Bomb', 'v':'Random Map Vote', 'V':'Map Vote', 'w':'Weapon Restrictions', 'W':'Freeze Bomb',
          'x':'Client Execution', 'X':'Adjust Health', 'y':'Client Execution Menu', 'Y':'Beacon Player', 'z':'Config', 'Z':'Give Player Item', '#':'All Flags'}
iflags = {'a':'Name Tag Kick Immunity', 'afk':'AFK Kick Immunity', 'autojoin':'Autojoin Immunity', 'b':'Ban Immunity', 'c':'Balance Team Immunity',
          'd':'Drug Immunity', 'e':'Burn Immunity', 'f':'Freeze Immunity', 'g':'Swap Immunity', 'grav':'Per Player Gravity Immunity', 'h':'Mute Immunity',
          'i':'Blind Immunity', 'immunity':'Basic Immunity', 'j':'Gimp Immunity', 'k':'Kick Immunity', 'l':'Slap Immunity', 'm':'Slay Immunity', 'n':'Reserved Slot Kick Immunity',
          'o':'Set Skin Immunity', 'p':'Reserved Skin Immunity', 'ping':'Ping Kick Immunity', 'q':'Timebomb Immunity', 'r':'Firebomb Immunity', 's':'Freezebomb Immunity',
          't':'Teleport Immunity', 'u':'Beacon Immunity', 'v':'Ghost Immunity', 'w':'Give Item Immunity', 'x':'Client Execution Immunity', 'y':'Color Change Immunity', '#':'All Flags'}
# ======================================================


# ======================================================
# ERROR CLASSES
# ===========================
class AddonError(Exception):
	pass
class FileError(Exception):
	pass
class GroupError(Exception):
    pass
class ClientError(Exception):
    pass
class FlagError(Exception):
    pass
class CommandError(Exception):
    pass

def ce(cmd):
    raise CommandError("Command %s is not included in manilib: Client command" % cmd)
# ======================================================


# ======================================================
# CUSTOM ADDONS
# ===========================
def registerCustomAddon(addonName):
    if isAddon(addonName):
        raise AddonError("Addon %s is already registered, get with getAddon(\"%s\")" % (addonName, addonName))
    
    dict_caddons[addonName] = {}
    dict_caddons[addonName] = Addon(str(addonName))
    return dict_caddons[addonName]

def unregisterCustomAddon(addonName):
    del dict_caddons[addonName]

def getAddon(addonName):
    if not isAddon(addonName):
        raise AddonError("Addon %s is not registered, register it with registerCustomAddon(\"%s\")" % (addonName, addonName))
    
    return dict_caddons[addonName]

class Addon(object):
    def __init__(self, name):
        self.name = str(name)
        self.dtitle = dict_caddons[name]['dtitle']
        self.attributes = dict_caddons[name]
    
    def __getitem__(self, item):
        item = str(item).lower()
        
        if item not in self.attributes:
            raise KeyError(item)
        
        return self.attributes[item]
    
    def __getattr__(self, item):
        item = str(item).lower()
        
        if item not in self.attributes:
            raise KeyError(item)
        
        return self.attributes[item]

    def setDisplayTitle(self, title):
        self.dtitle = title
        dict_caddons[self.name]['dtitle'] = title

    def getDisplayTitle(self):
        return self.dtitle
# ======================================================


# ======================================================
# MSG/TELL
# ===========================
def msg(addon, msg, mainprefix=1, addonprefix=1):
    if not isAddon(addon) and addon != "mani":
        raise AddonError("Addon %s is not registered" % addon)
    
    if addon == "mani":
        mainprefix = 1
        addonprefix = 0
    
    if mainprefix:
        if addonprefix:
            msg = "[#greenMANI - %s#default] %s" % (getAddon(addon).getDisplayTitle(), msg)
        else:
            msg = "[#greenMANI#default] %s" % msg
    else:
        if addonprefix:
            msg = "[#green%s#default] %s" % (getAddon(addon).getDisplayTitle(), msg)
    
    es.msg("#multi", msg)

def tell(userid, addon, msg, mainprefix=1, addonprefix=1):
    if not isAddon(addon) and addon != "mani":
        raise AddonError("Addon %s is not registered" % addon)
    
    if addon == "mani":
        mainprefix = 1
        addonprefix = 0
    
    if mainprefix:
        if addonprefix:
            msg = "[#greenMANI - %s#default] %s" % (getAddon(addon).getDisplayTitle(), msg)
        else:
            msg = "[#greenMANI#default] %s" % msg
    else:
        if addonprefix:
            msg = "[#green%s#default] %s" % (getAddon(addon).getDisplayTitle(), msg)
    
    es.tell(userid, "#multi", msg)
# ======================================================


# ======================================================
# READ CLIENTS
# ===========================
def readClients(): 
    if not os.path.isfile(clientsPath):
        raise FileError("Couldn't load clients.txt: No such file")

    global dict_mani
    # Create admins dictionary
    dict_mani['clients'] = {}
    dict_mani['groups'] = {'admin':[], 'immunity':[]}
    dict_mani['levels'] = {'admin':[], 'immunity':[]}

    # Open clients.txt
    file = open(clientsPath, 'r')
    # Read lines
    lines = file.readlines()
    # Close clients.txt
    file.close()

    # CorrectLines tuple
    cl = []

    # Go through every line in lines
    for line in lines:
        # Replace indents by nothing, space by nothing and "" by " " IF the line doesn't not start with "Admin" or "Immunity" (flags, spaces needed)
        line = line.replace('\t', '').replace('""', '" "')
        
        if not line.startswith('//') and not line.startswith('"clients.txt"') and not line.startswith('"version"') and not line == "\n" and not line.startswith('{')  and not line.startswith('}') and not line.startswith('"players"'):
            # Append the line to the CorrectLines tuple
            cl.append(line.replace('\n', ''))

    for line in cl:
        if line.startswith('"name" "'):
            startindex = cl.index(line)
            homei = startindex - 1
            flagsstart = homei + 3
            
            names = []
            steamids = []
            
            uniqueid = cl[homei].replace('"', '')
            
            while cl[startindex].startswith('"name" "'):
                names.append(cl[startindex].replace('"name" "', '').replace('"', ''))
                startindex += 1
            
            while cl[startindex].startswith('"steam" "'):
                steamids.append(cl[startindex].replace('"steam" "', '').replace('"', ''))
                startindex += 1

            # Insert ->
            dict_mani['clients'][uniqueid] = {'name':names, 'steam':steamids, 'uniqueid':uniqueid, 'aflags':[], 'iflags':[], 'cflags':[], 'agroups':[], 'igroups':[], 'alevels':[], 'ilevels':[]}

            # GET ADMIN AND IMMUNITY FLAGS
            if cl[flagsstart] == '"flags"':
                flagsindex = flagsstart + 1
                while cl[flagsindex].startswith('"Immunity" "'):
                    for flag in cl[flagsindex].replace('"Immunity" "', '').replace('"', '').split():
                        dict_mani['clients'][uniqueid]['iflags'].append(flag)
                    flagsindex += 1
                while cl[flagsindex].startswith('"Admin" "'):
                    for flag in cl[flagsindex].replace('"Admin" "', '').replace('"', '').split():
                        dict_mani['clients'][uniqueid]['aflags'].append(flag)
                    flagsindex += 1
                start = flagsindex
            else:
                start = flagsstart
            
            # GET ADMIN AND IMMUNITY GROUPS
            if cl[start] == '"groups"':
                groupindex = start + 1
                while cl[groupindex].startswith('"Admin" "'):
                    dict_mani['clients'][uniqueid]['agroups'].append(cl[groupindex].replace('"Admin" "', '').replace('"', ''))
                    groupindex += 1
                while cl[groupindex].startswith('"Immunity" "'):
                    dict_mani['clients'][uniqueid]['igroups'].append(cl[groupindex].replace('"Immunity" "', '').replace('"', ''))
                    groupindex += 1
                start = groupindex
            else:
                start = start

            # GET ADMIN AND IMMUNITY LEVELS
            if cl[start] == '"levels"':
                levelsindex = start + 1
                while cl[levelsindex].startswith('"Admin" "'):
                    dict_mani['clients'][uniqueid]['alevels'].append(cl[levelsindex].replace('"Admin" "', '').replace('"', ''))
                    levelsindex += 1
                while cl[levelsindex].startswith('"Immunity" "'):
                    dict_mani['clients'][uniqueid]['ilevels'].append(cl[levelsindex].replace('"Immunity" "', '').replace('"', ''))
                    levelsindex += 1
                start = levelsindex
    # GET GROUPS
    if cl[start] == '"groups"':
        groupindex = start + 1
        try:
            if cl[groupindex] == '"Immunity"':
                groupindex += 1
                while cl[groupindex] != '"Admin"':
                    groupname, flags = cl[groupindex].split('" "')
                    groupname = groupname.replace('"', '')
                    members = []
                    for client in dict_mani['clients']:
                        if groupname in dict_mani['clients'][client]['igroups']:
                            members.append(client)
                    dict_mani['groups']['immunity'][groupname] = {'flags':flags.replace('"', '').split(), 'members':members}
                    groupindex += 1
            if cl[groupindex] == '"Admin"':
                groupindex += 1
                while cl[groupindex] != '"levels"':
                    groupname, flags = cl[groupindex].split('" "')
                    groupname = groupname.replace('"', '')
                    members = []
                    for client in dict_mani['clients']:
                        if groupname in dict_mani['clients'][client]['agroups']:
                            members.append(client)
                    dict_mani['groups']['admin'][groupname] = {'flags':flags.replace('"', '').split(), 'members':members}
                    groupindex += 1
            groupstart = groupindex
        except:
            groupstart = groupindex
    if cl[groupstart] == '"levels"':
        groupindex = groupstart + 1
        try:
            if cl[groupindex] == '"Immunity"':
                groupindex += 1
                while cl[groupindex] != '"Admin"':
                    groupname, flags = cl[groupindex].split('" "')
                    groupname = groupname.replace('"', '')
                    members = []
                    for client in dict_mani['clients']:
                        if groupname in dict_mani['clients'][client]['ilevels']:
                            members.append(client)
                    dict_mani['levels']['immunity'][groupname] = {'flags':flags.replace('"', '').split(), 'members':members}
                    groupindex += 1
            if cl[groupindex] == '"Admin"':
                groupindex += 1
                try:
                    while cl[groupindex]:
                        groupname, flags = cl[groupindex].split('" "')
                        groupname.replace('"', '')
                        members = []
                        for client in dict_mani['clients']:
                            if groupname in dict_mani['clients'][client]['alevels']:
                                members.append(client)
                        dict_mani['levels']['admin'][groupname] = {'flags':flags.replace('"', '').split(), 'members':members}
                        groupindex += 1
                except:
                    pass
        except:
            pass
            
def reloadClients():
    # If the clients.txt was never read, do it
    try:
        if dict_mani['clients']:
            # Save all custom flags
            saveCustomFlags()
            # Clear up the dictionary
            del dict_mani['clients']
            del dict_mani['groups']
            del dict_mani['levels']
            # Read the clients.txt
            readClients()
            # Load all custom flags
            loadCustomFlags()
    except:
        readClients()
        loadCustomFlags()
# ======================================================


# ======================================================
# PICKLE CFLAGS
# ===========================
def loadCustomFlags():
    try:
        cflagsdb = open(cflagsPath, 'r')
        dumpclients = cPickle.load(cflagsdb)
        for client in dumpclients:
            dict_mani['clients'][client]['cflags'] = dumpclients[client]['cflags']
        cflagsdb.close()
    except:
        pass

def saveCustomFlags():
    cflagsdb = open(cflagsPath, 'w')
    dumpclients = {}
    for client in dict_mani['clients']:
        if dict_mani['clients'][client]['cflags']:
            dumpclients[client] = {'cflags':dict_mani['clients'][client]['cflags']}
    if dumpclients:
        cPickle.dump(dumpclients, cflagsdb)
    cflagsdb.close()
# ======================================================


# ======================================================
# CLIENTS CLASS
# ===========================
class Client(object):
    def __init__(self, id):
        self.id = str(id)
        if not isClient(self.id):
            raise ClientError("Couldn't get id (%s): No such client" % self.id)
        
        self.uniqueid = getUniqueId(self.id)
        self.name = dict_mani['clients'][self.uniqueid]['name']
        self.steam = dict_mani['clients'][self.uniqueid]['steam']
        self.aflags = dict_mani['clients'][self.uniqueid]['aflags']
        self.iflags = dict_mani['clients'][self.uniqueid]['iflags']
        self.agroups = dict_mani['clients'][self.uniqueid]['agroups']
        self.igroups = dict_mani['clients'][self.uniqueid]['igroups']
        self.alevels = dict_mani['clients'][self.uniqueid]['alevels']
        self.ilevels = dict_mani['clients'][self.uniqueid]['ilevels']
        self.attributes = dict_mani['clients'][self.uniqueid]
    
    def __getitem__(self, item):
        item = str(item).lower()
        
        if item not in self.attributes:
            raise KeyError(item)
        
        return self.attributes[item]
    
    def __getattr__(self, item):
        item = str(item).lower()
        
        if item not in self.attributes:
            raise KeyError(item)
        
        return self.attributes[item]
    
    def get(self, item):
        return self.__getitem__(item)
    
    def remove(self):
        client('RemoveClient "%s"' % self.uniqueid)
        del dict_mani['clients'][self.uniqueid]

    def addSteam(self, steamid):
        client('AddSteam "%s" "%s"' % (self.uniqueid, steamid))
        dict_mani['clients'][self.uniqueid]['steam'].append(steamid)
    
    def hasAFlag(self, flag):
        flag = str(flag)
        
        if not isAdminFlag(flag):
            raise FlagError("Couldn't check user (%s) for admin-flag %s: Invalid flag" % (self.name, flag))

        if flag in dict_mani['clients'][self.uniqueid]['aflags']:
            return 1
        else:
            return 0

    def hasIFlag(self, flag):
        flag = str(flag)
        
        if not isImmunityFlag(flag):
            raise FlagError("Couldn't check user (%s) for immunity-flag %s: Invalid flag" % (self.name, flag))

        if flag in dict_mani['clients'][self.uniqueid]['iflags']:
            return 1
        else:
            return 0
    
    def hasCFlag(self, flag):
        flag = str(flag)
        
        if flag in dict_mani['clients'][self.uniqueid]['cflags']:
            return 1
        else:
            return 0
        
    def inAGroup(self, group):
        group = str(group)

        if group in dict_mani['clients'][self.uniqueid]['agroups']:
            return 1
        else:
            return 0

    def inIGroup(self, group):
        group = str(group)

        if group in dict_mani['clients'][self.uniqueid]['igroups']:
            return 1
        else:
            return 0

    def hasALevel(self, level):
        level = str(level)

        if level in dict_mani['clients'][self.uniqueid]['alevels']:
            return 1
        else:
            return 0

    def hasILevel(self, level):
        level = str(level)

        if level in dict_mani['clients'][self.uniqueid]['ilevels']:
            return 1
        else:
            return 0
    
    def addAGroup(self, group):
        if not isAdminGroup(group):
            raise GroupError("Couldn't add user (%s) to group %s: No such group" % (self.uniqueid, group))
        elif self.inAGroup(group):
            raise GroupError("Couldn't add user (%s) to group %s: Already in group" % (self.uniqueid, group))
        client('AddAGroup "%s" "%s"' % (self.uniqueid, group))
        dict_mani['clients'][self.uniqueid]['agroups'].append(group)
    
    def addIGroup(self, group):
        if not isImmunityGroup(group):
            raise GroupError("Couldn't add user (%s) to group %s: No such group" % (self.uniqueid, group))
        elif self.inIGroup(group):
            raise GroupError("Couldn't add user (%s) to group %s: Already in group" % (self.uniqueid, group))
        client('AddIGroup "%s" "%s"' % (self.uniqueid, group))
        dict_mani['clients'][self.uniqueid]['igroups'].append(group)
    
    def removeAGroup(self, group):
        if not isAdminGroup(group):
            raise GroupError("Couldn't add user (%s) from group %s: No such group" % (self.uniqueid, group))
        elif not self.inAGroup(group):
            raise GroupError("Couldn't add user (%s) from group %s: Not in group" % (self.uniqueid, group))
        client('RemoveAGroup "%s" "%s"' % (self.uniqueid, group))
        dict_mani['clients'][self.uniqueid]['agroups'].remove(group)
    
    def removeIGroup(self, group):
        if not isImmunityGroup(group):
            raise GroupError("Couldn't add user (%s) from group %s: No such group" % (self.uniqueid, group))
        elif not self.inIGroup(group):
            raise GroupError("Couldn't add user (%s) from group %s: Not in group" % (self.uniqueid, group))
        client('RemoveIGroup "%s" "%s"' % (self.uniqueid, group))
        dict_mani['clients'][self.uniqueid]['igroups'].remove(group)
    
    def addAFlag(self, flag):
        if not isAdminFlag(flag):
            raise FlagError("Couldn't add flag (%s) to user (%s): No such flag" % (flag, self.uniqueid))
        elif self.hasAFlag(flag):
            raise FlagError("Couldn't add flag (%s) to user (%s): Already has flag" % (flag, self.uniqueid))
        client('SetAFlag "%s" "+%s"' % (self.uniqueid, flag))
        if flag == '#':
            for flag in aflags:
                if not flag == '#':
                    dict_mani['clients'][self.uniqueid]['aflags'].append(flag)
        else:
            dict_mani['clients'][self.uniqueid]['aflags'].append(flag)
    
    def removeAFlag(self, flag):
        if not isAdminFlag(flag):
            raise FlagError("Couldn't remove flag (%s) from user (%s): No such flag" % (flag, self.uniqueid))
        elif not self.hasAFlag(flag):
            raise FlagError("Couldn't remove flag (%s) from user (%s): Flag not added" % (flag, self.uniqueid))
        client('SetAFlag "%s" "-%s"' % (self.uniqueid, flag))
        if flags == '#':
            for flag in aflags:
                if not flag == '#':
                    dict_mani['clients'][self.uniqueid]['aflags'].remove(flag)
        else:
            dict_mani['clients'][self.uniqueid]['aflags'].remove(flag)
    
    def addIFlag(self, flag):
        if not isImmuntyFlag(flag):
            raise FlagError("Couldn't add flag (%s) to user (%s): No such flag" % (flag, self.uniqueid))
        elif self.hasIFlag(flag):
            raise FlagError("Couldn't add flag (%s) to user (%s): Already has flag" % (flag, self.uniqueid))
        client('SetIFlag "%s" "+%s"' % (self.uniqueid, flag))
        if flag == '#':
            for flag in iflags:
                if not flag == '#':
                    dict_mani['clients'][self.uniqueid]['ilags'].append(flag)
        else:
            dict_mani['clients'][self.uniqueid]['iflags'].append(flag)
    
    def removeIFlag(self, flag):
        if not isImmuntyFlag(flag):
            raise FlagError("Couldn't remove flag (%s) from user (%s): No such flag" % (flag, self.uniqueid))
        elif not self.hasAFlag(flag):
            raise FlagError("Couldn't remove flag (%s) from user (%s): Flag not added" % (flag, self.uniqueid))
        client('SetIFlag "%s" "-%s"' % (self.uniqueid, flag))
        if flags == '#':
            for flag in iflags:
                if not flag == '#':
                    dict_mani['clients'][self.uniqueid]['iflags'].remove(flag)
        else:
            dict_mani['clients'][self.uniqueid]['iflags'].remove(flag)
    
    def addCFlag(self, flag):
        if self.hasCFlag(flag):
            raise FlagError("Couldn't add flag (%s) to user (%s): Already has flag" % (flag, self.uniqueid))
        dict_mani['clients'][self.uniqueid]['cflags'].append(flag)
        saveCustomFlags()
    
    def removeCFlag(self, flag):
        if not self.hasCFlag(flag):
            raise FlagError("Couldn't remove flag (%s) from user (%s): Flag not added" % (flag, self.uniqueid))
        dict_mani['clients'][self.uniqueid]['cflags'].remove(flag)
        saveCustomFlags()
# ======================================================


# ======================================================
# GROUP CLASS
# ===========================
class Group(object):
    def __init__(self, name, filter):
        self.name = str(name)
        self.filter = str(filter)
        self.type = None
        
        if self.filter == 'a':
            self.type = 'admin'
        elif self.filter == 'i':
            self.type = 'immunity'
        
        if not isGroup(self.name, self.filter):
            raise GroupError("%s is not a valid group" % self.name)
        else:
            self.attributes = dict_mani['groups'][self.type][self.name]
            self.flags = dict_mani['groups'][self.type][self.name]['flags']
            self.members = dict_mani['groups'][self.type][self.name]['members']
    
    def __getitem__(self, item):
        item = str(item).lower()
        
        if item not in self.attributes:
            raise KeyError(item)
        
        return self.attributes[item]
    
    def __getattr__(self, item):
        item = str(item).lower()
        
        if item not in self.attributes:
            raise KeyError(item)
        
        return self.attributes[item]
    
    def get(self, item):
        return self.__getitem__(item)
    
    def remove(self):
        clientgroup('remove%sgroup "%s"' % (self.filter, self.name))
        del dict_mani['groups'][self.type][self.name]
    
    def hasFlag(self, flag):
        if not isFlag(flag, self.filter):
            raise FlagError("Couldn't check group (%s) for flag %s: Invalid flag" % (self.name, flag))
        if flag in dict_mani['groups'][self.type][self.name]['flags']:
            return 1
        else:
            return 0
    
    def addFlag(self, flag):
        if not isFlag(flag, self.filter):
            raise FlagError("Couldn't add flag (%s) to group (%s): No such flag" % (flag, self.name))
        elif self.hasFlag(flag):
            raise FlagError("Couldn't add flag (%s) to group (%s): Already has flag" % (flag, self.name))
        if flag == '#':
            if self.filter == 'a':
                for aflag in aflags:
                    dict_mani['groups'][self.type][self.name]['flags'].append(aflag)
            elif self.filter == 'i':
                for iflag in iflags:
                    dict_mani['groups'][self.type][self.name]['flags'].append(iflag)
            else:
                dict_mani['groups'][self.type][self.name]['flags'].append(flag)
        clientgroup('add%sgroup "%s" "+%s"' % (self.filter, self.name, flag))
    
    def removeFlag(self, flag):
        if not isFlag(flag, self.filter):
            raise FlagError("Couldn't remove flag (%s) from group (%s): No such flag" % (flag, self.name))
        elif not self.hasFlag(flag):
            raise FlagError("Couldn't remove flag (%s) from group (%s): Flag not added" % (flag, self.name))
        if flag == '#':
            if self.filter == 'a':
                for aflag in aflags:
                    dict_mani['groups'][self.type][self.name]['flags'].remove(aflag)
            elif self.filter == 'i':
                for iflag in iflags:
                    dict_mani['groups'][self.type][self.name]['flags'].remove(iflag)
            else:
                dict_mani['groups'][self.type][self.name]['flags'].append(flag)
            
        clientgroup('add%sgroup "%s" "-%s"' % (self.filter, self.name, flag))
# ======================================================


# ======================================================
# LEVEL CLASS
# ===========================
class Level(object):
    def __init(self, level, filter):
        self.level = str(level)
        self.filter = str(filter)
        
        if self.filter == 'a':
            self.type = 'admin'
        elif self.filter == 'i':
            self.type = 'immunity'
        
        if not isLevel(self.level, self.filter):
            raise LevelError("%s is not a valid level" % self.level)
        
        self.attributes = dict_mani['levels'][self.type][self.level]
        self.flags = dict_mani['levels'][self.type][self.level]['flags']
        self.members = dict_mani['levels'][self.type][self.level]['members']
    
    def __getitem__(self, item):
        item = str(item).lower()
        
        if item not in self.attributes:
            raise KeyError(item)
        
        return self.attributes[item]

    def __getattr__(self, item):
        item = str(item).lower()
        
        if item not in self.attributes:
            raise KeyError(item)
        
        return self.attributes[item]
    
    def get(self, item):
        return self.__getitem__(item)
    
    def remove(self):
        clientgroup('remove%slevel "%s"' % (self.filter, self.level))
        del dict_mani['levels'][self.type][self.level]

    def hasFlag(self, flag):
        if not isFlag(flag, self.filter):
            raise FlagError("Couldn't check level (%s) for flag %s: Invalid flag" % (self.id, flag))
        if flag in dict_mani['levels'][self.type][self.id]['flags']:
            return 1
        else:
            return 0
    
    def addFlag(self, flag):
        if not isFlag(flag, self.filter):
            raise FlagError("Couldn't add flag (%s) to level (%s): No such flag" % (flag, self.id))
        clientgroup('add%slevel "%s" "+%s"' % (self.filter, self.id, flag))
        dict_mani['levels'][self.type][self.id]['flags'].append(flag)
    
    def removeFlag(self, flag):
        if not isFlag(flag, self.filter):
            raise FlagError("Couldn't remove flag (%s) from level (%s): No such flag" % (flag, self.id))
        clientgroup('add%slevel "%s" "-%s"' % (self.filter, self.id, flag))
        dict_mani['levels'][self.type][self.id]['flags'].remove(flag)

# ======================================================


# ======================================================
# PLAYER
# ===========================
class Player(object):
    def __init__(self, userid):
        if not es.exists('userid', userid):
            raise ClientError("No such userid: %s" % userid)
        self.userid = str(userid)
        self.target = es.getplayername(userid)
        self.steamid = es.getplayersteamid(userid)
        
    def __getitem__(self, item):
        item = str(item).lower()
        
        if item not in self.attributes:
            raise KeyError(item)
        
        return self.attributes[item]
        
    def __getattr__(self, item):
        item = str(item).lower()
        
        if item not in self.attributes:
            raise KeyError(item)
        
        return self.attributes[item]
    
    def ban(self, time):
        cmd('ban', self.steamid, time)
    
    def kick(self):
        cmd('kick', self.userid)
    
    def beacon(self, mode=None):
        if mode:
            cmd('beacon', self.target, mode)
            return
        cmd('beacon', self.target)
    
    def blind(self, amount=None):
        if amount:
            cmd('blind', self.target, amount)
            return
        cmd('blind', self.target)
    
    def burn(self, mode=None):
        cmd('burn', self.target)
    
    def cexec(self, command):
        cmd('cexec', self.target, command)
    
    def color(red, blue, green, alpha):
        cmd('color', self.target, '%s %s %s %s' % (str(red), str(blue), str(green), str(alpha)))
    
    def color_weapon(red, blue, green, alpha):
        cmd('color_weapon', self.target, '%s %s %s %s' % (str(red), str(blue), str(green), str(alpha)))
    
    def colour(red, blue, green, alpha):
        cmd('colour', self.target, '%s %s %s %s' % (str(red), str(blue), str(green), str(alpha)))
    
    def colour_weapon(red, blue, green, alpha):
        cmd('colour_weapon', self.target, '%s %s %s %s' % (str(red), str(blue), str(green), str(alpha)))
    
    def drug(self, mode=None):
        if mode:
            cmd('drug', self.target, mode)
            return
        cmd('drug', self.target)
    
    def firebomb(self, mode=None):
        if mode:
            cmd('firebomb', self.target, mode)
            return
        cmd('firebomb', self.target)

    def freeze(self, mode=None):
        if mode:
            cmd('freeze', self.target, mode)
            return
        cmd('freeze', self.target)
    
    def freezebomb(self, mode=None):
        if mode:
            cmd('freezebomb', self.target, mode)
            return
        cmd('freezebomb', self.target)
    
    def gimp(self, mode=None):
        if mode:
            cmd('gimp', self.target, mode)
            return
        cmd('gimp', self.target)
    
    def give(self, item):
        cmd('give', self.target, item)

    def give_ammo(self, slot, pfmode, amount, noise=0):
        cmd('give_ammo', self.target, '%s %s %s %s %s' % (str(slot), str(pfmode), str(amount), str(noise)))
    
    def givecash(self, amount):
        cmd('givecash', self.target, amount)
    
    def givecashp(self, percent):
        cmd('givecashp', self.target, percent)
    
    def givehealth(self, amount):
        cmd('givehealth', self.target, amount)
    
    def givehealthp(self, percent):
        cmd('givehealthp', self.target, percent)
    
    def gravity(self, percent):
        cmd('gravity', self.target, percent)
    
    def mute(self, mode):
        if mode:
            cmd('mute', self.target, mode)
            return
        cmd('mute', self.target)
    
    def msay(self, time, msg):
        cmd('msay', None, '%s "%s" %s' % (time, self.target, msg))
    
    def noclip(self, mode):
        if mode:
            cmd('mute', self.target, mode)
            return
        cmd('mute', self.target)
    
    def psay(self, msg):
        cmd('psay', self.target, msg)
    
    def render_fx(self, type):
        cmd('render_fx', self.target, type)
    
    def render_mode(self, mode):
        cmd('render_mode', self.target, mode)
    
    def resetrank(self):
        cmd('resetrank', self.steamid)
    
    def setcash(self, amount):
        cmd('setcash', self.target, amount)
    
    def sethealth(self, amount):
        cmd('sethealth', self.target, amount)
    
    def setskin(self, skin):
        cmd('setskin', self.target, skin)
    
    def slap(self, damage=0):
        cmd('slap', self.target, damage)
    
    def slay(self):
        cmd('slay', self.target)
    
    def spec(self):
        cmd('spec', self.target)
    
    def swapteam(self):
        cmd('swapteam', self.target)
    
    def swapteam_d(self):
        cmd('swapteam_d', self.target)
    
    def takecash(self, amount):
        cmd('takecash', self.target, amount)
    
    def takecashp(self, percent):
        cmd('takecashp', self.target, percent)
    
    def takehealth(self, amount):
        cmd('takehealth', self.target, amount)
    
    def takehealthp(self, percent):
        cmd('takehealthp', self.target, percent)
    
    def teamindex(self):
        ce("teamindex")
    
    def teleport(self, x=None, y=None, z=None):
        if x and y and z:
            cmd('teleport', self.target, '%s %s %s' % (str(x), str(y), str(z)))
            return
        cmd('teleport', self.target)
    
    def timebomb(self, mode=None):
        if mode:
            cmd('timebomb', self.target, mode)
            return
        cmd('timebomb', self.target)
# ======================================================


# ======================================================
# CHECKS 'N' GETS
# ===========================
def isClient(uniqueid):
    return uniqueid in dict_mani['clients'] or getUniqueId(uniqueid) in dict_mani['clients']

def isAddon(addon):
    return addon in dict_caddons

def isAdminGroup(name):
    return name in dict_mani['groups']['admin']

def isImmunityGroup(name):
    return name in dict_mani['groups']['immunity']

def isAdminLevel(name):
    return name in dict_mani['levels']['admin']

def isImmunityLevel(name):
    return name in dict_mani['levels']['immunity']

def isAdminFlag(flag):
    return flag in aflags

def isImmunityFlag(flag):
    return flag in iflags

def isFlag(flag, filter):
    if filter == 'a':
        return isAdminFlag(flag)
    elif filter == 'i':
        return isImmunityFlag(flag)

def getUniqueId(id):
    # If the id does not start with STEAM_, it can be the userid
    if not id.upper().startswith("STEAM_"):
        id = es.getplayersteamid(id)
    for client in dict_mani['clients']:
        if id.upper() in dict_mani['clients'][client]['steam']:
            return client
# ======================================================


# ======================================================
# MANI COMMANDS
# ===========================
def cmd(cmd, target=None, subcommand=None):
    if target != None:
        if subcommand != None:
            command = 'ma_%s "%s" %s' % (str(cmd), str(target), str(subcommand))
        else:
            command = 'ma_%s "%s"' % (str(cmd), str(target))
    else:
        if subcommand != None:
            command = 'ma_%s %s' % (str(cmd), str(subcommand))
        else:
            command = "ma_%s" % str(cmd)
    es.server.queuecmd(command)
# =============
# ma_client
# =============
def client(subcommand):
    cmd("client", subcommand)

def addClient(id, steamid=None):
    client('AddClient "%s"' % id)
    reloadClients()
    if steamid:
        getPlayer(id).addSteam(steamid.upper()) 
# =============
# ma_clientgroup
# =============
def clientgroup(subcommand):
    cmd("clientgroup", subcommand)

def addAGroup(name):
    clientgroup('addagroup "%s" ""' % name)
    dict_mani['groups']['admin'][name] = {'flags':[], 'members':[]}
    
def addIGroup(name):
    clientgroup('addigroup "%s" ""' % name)
    dict_mani['groups']['immunity'][name] = {'flags':[], 'members':[]}

def addALevel(name):
    clientgroup('addalevel "%s" ""' % name)
    dict_mani['levels']['admin'][name] = {'flags':[], 'members':[]}
    
def addILevel(name):
    clientgroup('addilevel "%s" ""' % name)
    dict_mani['levels']['immunity'][name] = {'flags':[], 'members':[]}
# =============
# ma_aban_name
# =============
def aban_name(name, time=None):
    if time != None:
        cmd('aban_name', name, time)
    else:
        cmd('aban_name', name)
# =============
# ma_aban_pname
# =============
def aban_pname(name, time=None):
    if time != None:
        cmd('aban_pname', name, time)
    else:
        cmd('aban_pname', name)
# =============
# ma_admins
# =============
def admins():
    ce("admin")
# =============
# ma_admingroups
# =============
def admingroups():
    ce("admingroups")
# =============
# ma_akick_ip
# =============
def akick_ip(ip):
    cmd('akick_ip', ip)
# =============
# ma_akick_name
# =============
def akick_name(name):
    cmd('akick_name', name)
# =============
# ma_akick_pname
# =============
def akick_pname(pname):
    cmd('akick_pname', pname)
# =============
# ma_akick_steam
# =============
def akick_steam(steam):
    cmd('akick_steam', str(steam).upper())
# =============
# ma_ashow_ip
# =============
def ashow_ip():
    ce("ashow_ip")
# =============
# ma_ashow_name
# =============
def ashow_name():
    ce("ashow_name")
# =============
# ma_ashow_pname
# =============
def ashow_pname():
    ce("ashow_pname")
# =============
# ma_ashow_steam
# =============
def ashow_steam():
    ce("ashow_steam")
# =============
# ma_balance
# =============
def balance():
    cmd('balance')
# =============
# ma_banip
# =============
def banip(ip, time=None):
    if time != None:
        cmd('banip', ip, time)
    else:
        cmd('banip', ip)
# =============
# ma_browse
# =============
def browse(url=None):
    ce("browse")
# =============
# ma_cexec
# =============
class cexec(object):
    def __init__(self, command):
        self.command = '"%s"' % str(command)
    
    def user(self, target):
        cmd('cexec', target, self.command)
    
    def ct(self):
        cmd('cexec_ct', None, self.command)
    
    def t(self):
        cmd('cexec_t', None, self.command)
    
    def spec(self):
        cmd('cexec_spec', None, self.command)
    
    def all(self):
        cmd('cexec_all', None, self.command)
# =============
# ma_chat
# =============
def chat(msg):
    cmd('chat', None, msg)
# =============
# ma_chattrigers
# =============
def chattrigers():
    ce("chattrigers")
# =============
# ma_config
# =============
def config(cvar=None):
    ce("config")
# =============
# ma_csay
# =============
def csay(msg):
    cmd('csay', msg)
# =============
# ma_decal
# =============
def decal(decal):
    ce('decal')
# =============
# ma_dropc4
# =============
def dropc4():
    cmd('dropc4')
# =============
# ma_effect
# =============
def effect(todo):
    # TODO
    pass
# =============
# ma_explode
# =============
def explode():
    ce('explode')
# =============
# ma_endofobserve
# =============
def endofobserve():
    ce("endofobserve")
# =============
# ma_exit
# =============
def exit():
    ce("exit")
# =============
# ma_forcegametype
# =============
def forcegametype():
    cmd('forcegametype')
# =============
# ma_game
# =============
def game():
    ce('game')
# =============
# ma_help
# =============
def help(pattern=None):
    ce("help")
# =============
# ma_immunity
# =============
def immunity(groups=None):
    ce('immunity')
# =============
# ma_knives
# =============
def knives():
    cmd('knives')
# =============
# ma_listmaps
# =============
def listmaps():
    ce("listmaps")
# =============
# ma_map
# =============
def map(map):
    cmd('map', None, map)
# =============
# ma_mapcycle
# =============
def mapcycle():
    ce("mapcycle")
# =============
# ma_maphistory
# =============
def maphistory():
    ce("maphistory")
# =============
# ma_maplist
# =============
def maplist():
    ce("maplist")
# =============
# ma_msay
# =============
def msay(target, dtime, msg):
    cmd('msay', None, '%s "%s" "%s"' % (str(dtime), str(target), msg))
# =============
# ma_nextmap
# =============
def nextmap():
    ce("nextmap")
# =============
# ma_nosnipers
# =============
def nosnipers():
    cmd('nosnipers')
# =============
# ma_observe
# =============
def observe():
    ce("observe")
# =============
# ma_offset
# =============
def offset(offset):
    cmd('offset', None, offset)
# =============
# ma_offsetscan
# =============
def offsetscan(value, start, end):
    cmd('offsetscan', None, '%s %s %s' % (str(value), str(start), str(end)))
# =============
# ma_pistols
# =============
def pistols():
    cmd('pistols')
# =============
# ma_play
# =============
def play(index):
    cmd('play', None, index)
# =============
# ma_plranks
# =============
def plranks(start=None, end=None):
    ce("plranks")
# =============
# ma_pmess
# =============
def pmess(target=None):
    ce("pmess")
# =============
# ma_position
# =============
def position():
    ce("position")
# =============
# ma_ranks
# =============
def ranks(start=None, stop=None):
    ce("ranks")
# =============
# ma_rates
# =============
def rates():
    ce("rates")
# =============
# ma_rcon
# =============
def rcon(cmd):
    ce('rcon')
# =============
# ma_reloadclients
# =============
def reloadclients():
    cmd('reloadclients')
    reloadClients()
# =============
# ma_restrict
# =============
def restrict(weapon, amount=None):
    if amount:
        cmd('restrict', None, '%s %s' % (weapon, str(amount)))
        return
    cmd('restrict', None, weapon)
# =============
# ma_restrictall
# =============
def restrictall():
    cmd('restrictall')
# =============
# ma_restrictratio
# =============
def restrictratio(weapon, ratio):
    cmd('restrictratio', None, '%s %s' % (weapon, str(ratio)))
# =============
# ma_saveloc
# =============
def saveloc():
    ce("saveloc")
# =============
# ma_say
# =============
def say(msg):
    cmd('say', None, msg)
# =============
# ma_session
# =============
def session(target=None):
    ce("session")
# =============
# ma_setnextmap
# =============
def setnextmap(map):
    cmd('setnextmap', None, map)
# =============
# ma_shotguns
# =============
def shotguns():
    cmd('shotguns')
# =============
# ma_showrestrict
# =============
def showrestrict():
    ce("showrestrict")
# =============
# ma_showsounds
# =============
def showsounds():
    ce("showsounds")
# =============
# ma_skipmap
# =============
def skipmap():
    cmd('skipmap')
# =============
# ma_spray
# =============
def spray():
    ce("spray")
# =============
# ma_statsme
# =============
def statsme(ot=None):
    ce("statsme")
# =============
# ma_timeleft
# =============
def timeleft():
    ce("timeleft")
# =============
# ma_tklist
# =============
def tklist():
    ce("tklist")
# =============
# ma_unaban_name
# =============
def unauto_name(name):
    cmd('unaban_name', name)
# =============
# ma_unaban_pname
# =============
def unaban_pname(pname):
    cmd('unaban_pname', pname)
# =============
# ma_unakick_ip
# =============
def unakick_ip(ip):
    cmd('unakick_ip', ip)
# =============
# ma_unakick_name
# =============
def unakick_name(name):
    cmd('unakick_name', name)
# =============
# ma_unakick_pname
# =============
def unakick_pname(pname):
    cmd('unakick_pname', pname)
# =============
# ma_unakick_steam
# =============
def unakick_steam(steam):
    cmd('unakick_steam', None, steam)
# =============
# ma_unban
# =============
def unban(steam):
    cmd('unban', None, steam)
# =============
# ma_unrestrict
# =============
def unrestrict(weapon):
    cmd('unrestrict', None, weapon)
# =============
# ma_unrestrictall
# =============
def unrestrictall():
    cmd('unrestrictall')
# =============
# ma_users
# =============
def users():
    ce("users")
# =============
# ma_vote
# =============
def vote(maps, delay=None):
    if delay != None:
        command = delay
    
    if type(maps).__name__ == "str":
        try:
            if command:
                command = "%s %s" % (command, maps)
        except:
            command = maps
    else:
        for map in maps:
            try:
                if command:
                    command = "%s %s" % (command, map)
            except:
                command = map
    cmd('vote', None, command)
# =============
# ma_votecancel
# =============
def votecancel():
    cmd('votecancel')
# =============
# ma_voteextend
# =============
def voteextend():
    cmd('voteextend')
# =============
# ma_votemaplist
# =============
def votemaplist():
    ce("votemaplist")
# =============
# ma_votequestion
# =============
def votequestion(title, answers):
    command = '"%s"' % title
    if type(answers).__name__ == "str":
        command = '%s "%s"' % (command, answers)
    else:
        for answer in answers:
            command = '%s "%s"' % (command, answer)
    cmd('votequestion', None, command)
# =============
# ma_voterandom
# =============
def voterandom(delaytype, mapscount):
    cmd('voterandom', None, "%s %s" % (delaytype, mapscount))
# =============
# ma_votercon
# =============
def votercon(title, cmd):
    cmd('votercon', None, '"%s" "%s"' % (title, cmd))
# =============
# ma_war
# =============
def war(mode=None):
    if mode != None:
        cmd('war', None, mode)
    else:
        cmd('war')
# =============
# ======================================================


# ======================================================
# FUNCTIONS
# ===========================
def getClient(uniqueid):
    if not isClient(uniqueid):
        raise ClientError("Couldn't get player by id %s: No such user" % uniqueid)
    return Client(uniqueid)

def getPlayer(userid):
    return Player(userid)

def getAdminGroup(name):
    return Group(name, 'a')

def getImmunityGroup(name):
    return Group(name, 'i')

def getAddon(addon):
    if not isAddon(addon):
        raise AddonError("Couldn't get addon %s: No such addon" % addon)
    
    return Addon(addon)
# ======================================================